package com.sun.xml.internal.bind.v2.model.impl;

import com.sun.xml.internal.bind.v2.runtime.Name;
import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import com.sun.xml.internal.bind.v2.ClassFactory;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.UnmarshallingContext;
import com.sun.istack.internal.NotNull;
import org.xml.sax.SAXException;
import com.sun.xml.internal.bind.api.AccessorException;
import com.sun.xml.internal.bind.v2.runtime.XMLSerializer;
import com.sun.xml.internal.bind.v2.runtime.Location;
import java.lang.annotation.Annotation;
import com.sun.xml.internal.bind.v2.model.core.ClassInfo;
import com.sun.xml.internal.bind.v2.model.core.PropertyInfo;
import com.sun.xml.internal.bind.annotation.XmlLocation;
import javax.xml.bind.JAXBException;
import com.sun.xml.internal.bind.InternalAccessorFactory;
import java.lang.reflect.Modifier;
import java.util.Iterator;
import com.sun.xml.internal.bind.v2.runtime.reflect.TransducedAccessor;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeValuePropertyInfo;
import com.sun.xml.internal.bind.v2.model.core.PropertyKind;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimePropertyInfo;
import java.util.List;
import com.sun.xml.internal.bind.XmlAccessorFactory;
import com.sun.xml.internal.bind.v2.runtime.JAXBContextImpl;
import com.sun.xml.internal.bind.AccessorFactoryImpl;
import com.sun.xml.internal.bind.v2.runtime.IllegalAnnotationException;
import com.sun.xml.internal.bind.v2.model.annotation.Locatable;
import com.sun.xml.internal.bind.v2.runtime.Transducer;
import javax.xml.namespace.QName;
import java.util.Map;
import com.sun.xml.internal.bind.AccessorFactory;
import org.xml.sax.Locator;
import com.sun.xml.internal.bind.v2.runtime.reflect.Accessor;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeElement;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeClassInfo;
import java.lang.reflect.Method;
import java.lang.reflect.Field;
import java.lang.reflect.Type;

class RuntimeClassInfoImpl extends ClassInfoImpl<Type, Class, Field, Method> implements RuntimeClassInfo, RuntimeElement
{
    private Accessor<?, Locator> xmlLocationAccessor;
    private AccessorFactory accessorFactory;
    private boolean supressAccessorWarnings;
    private Accessor<?, Map<QName, String>> attributeWildcardAccessor;
    private boolean computedTransducer;
    private Transducer xducer;
    
    public RuntimeClassInfoImpl(RuntimeModelBuilder modelBuilder, Locatable upstream, Class clazz) {
        super(modelBuilder, upstream, clazz);
        this.supressAccessorWarnings = false;
        this.computedTransducer = false;
        this.xducer = null;
        this.accessorFactory = this.createAccessorFactory(clazz);
    }
    
    protected AccessorFactory createAccessorFactory(Class clazz) {
        AccessorFactory accFactory = null;
        JAXBContextImpl context = ((RuntimeModelBuilder)this.builder).context;
        if (context != null) {
            this.supressAccessorWarnings = context.supressAccessorWarnings;
            if (context.xmlAccessorFactorySupport) {
                XmlAccessorFactory factoryAnn = this.findXmlAccessorFactoryAnnotation(clazz);
                if (factoryAnn != null) {
                    try {
                        accFactory = (AccessorFactory)factoryAnn.value().newInstance();
                    }
                    catch (InstantiationException e) {
                        this.builder.reportError(new IllegalAnnotationException(Messages.ACCESSORFACTORY_INSTANTIATION_EXCEPTION.format(new Object[] { factoryAnn.getClass().getName(), this.nav().getClassName(clazz) }), this));
                    }
                    catch (IllegalAccessException e2) {
                        this.builder.reportError(new IllegalAnnotationException(Messages.ACCESSORFACTORY_ACCESS_EXCEPTION.format(new Object[] { factoryAnn.getClass().getName(), this.nav().getClassName(clazz) }), this));
                    }
                }
            }
        }
        if (accFactory == null) {
            accFactory = AccessorFactoryImpl.getInstance();
            return accFactory;
        }
        return accFactory;
    }
    
    protected XmlAccessorFactory findXmlAccessorFactoryAnnotation(Class clazz) {
        XmlAccessorFactory factoryAnn = (XmlAccessorFactory)this.reader().getClassAnnotation(XmlAccessorFactory.class, clazz, this);
        if (factoryAnn == null) {
            factoryAnn = (XmlAccessorFactory)this.reader().getPackageAnnotation(XmlAccessorFactory.class, clazz, this);
            return factoryAnn;
        }
        return factoryAnn;
    }
    
    @Override
    public Method getFactoryMethod() {
        return super.getFactoryMethod();
    }
    
    @Override
    public final RuntimeClassInfoImpl getBaseClass() {
        return (RuntimeClassInfoImpl)super.getBaseClass();
    }
    
    @Override
    protected ReferencePropertyInfoImpl createReferenceProperty(PropertySeed<Type, Class, Field, Method> seed) {
        return new RuntimeReferencePropertyInfoImpl(this, seed);
    }
    
    @Override
    protected AttributePropertyInfoImpl createAttributeProperty(PropertySeed<Type, Class, Field, Method> seed) {
        return new RuntimeAttributePropertyInfoImpl(this, seed);
    }
    
    @Override
    protected ValuePropertyInfoImpl createValueProperty(PropertySeed<Type, Class, Field, Method> seed) {
        return new RuntimeValuePropertyInfoImpl(this, seed);
    }
    
    @Override
    protected ElementPropertyInfoImpl createElementProperty(PropertySeed<Type, Class, Field, Method> seed) {
        return new RuntimeElementPropertyInfoImpl(this, seed);
    }
    
    @Override
    protected MapPropertyInfoImpl createMapProperty(PropertySeed<Type, Class, Field, Method> seed) {
        return new RuntimeMapPropertyInfoImpl(this, seed);
    }
    
    @Override
    public List<? extends RuntimePropertyInfo> getProperties() {
        return super.getProperties();
    }
    
    public RuntimePropertyInfo getProperty(String name) {
        return (RuntimePropertyInfo)super.getProperty(name);
    }
    
    public void link() {
        this.getTransducer();
        super.link();
    }
    
    public <B> Accessor<B, Map<QName, String>> getAttributeWildcard() {
        RuntimeClassInfoImpl c = this;
        while (c != null) {
            if (c.attributeWildcard != null) {
                if (c.attributeWildcardAccessor == null) {
                    c.attributeWildcardAccessor = c.createAttributeWildcardAccessor();
                }
                return c.attributeWildcardAccessor;
            }
            else {
                c = c.getBaseClass();
            }
        }
        return null;
    }
    
    public Transducer getTransducer() {
        if (!this.computedTransducer) {
            this.computedTransducer = true;
            this.xducer = this.calcTransducer();
        }
        return this.xducer;
    }
    
    private Transducer calcTransducer() {
        RuntimeValuePropertyInfo valuep = null;
        if (this.hasAttributeWildcard()) {
            return null;
        }
        for (RuntimeClassInfoImpl ci = this; ci != null; ci = ci.getBaseClass()) {
            for (RuntimePropertyInfo pi : ci.getProperties()) {
                if (pi.kind() == PropertyKind.VALUE) {
                    valuep = (RuntimeValuePropertyInfo)pi;
                }
                else {
                    return null;
                }
            }
        }
        if (valuep == null) {
            return null;
        }
        if (!valuep.getTarget().isSimpleType()) {
            return null;
        }
        return new RuntimeClassInfoImpl.TransducerImpl((Class)this.getClazz(), TransducedAccessor.get(((RuntimeModelBuilder)this.builder).context, valuep));
    }
    
    private Accessor<?, Map<QName, String>> createAttributeWildcardAccessor() {
        assert this.attributeWildcard != null;
        return ((RuntimeClassInfoImpl.RuntimePropertySeed)this.attributeWildcard).getAccessor();
    }
    
    @Override
    protected RuntimeClassInfoImpl.RuntimePropertySeed createFieldSeed(Field field) {
        boolean readOnly = Modifier.isStatic(field.getModifiers());
        Accessor acc;
        try {
            if (this.supressAccessorWarnings) {
                acc = ((InternalAccessorFactory)this.accessorFactory).createFieldAccessor((Class)this.clazz, field, readOnly, this.supressAccessorWarnings);
            }
            else {
                acc = this.accessorFactory.createFieldAccessor((Class)this.clazz, field, readOnly);
            }
        }
        catch (JAXBException e) {
            this.builder.reportError(new IllegalAnnotationException(Messages.CUSTOM_ACCESSORFACTORY_FIELD_ERROR.format(new Object[] { this.nav().getClassName(this.clazz), e.toString() }), this));
            acc = Accessor.getErrorInstance();
        }
        return new RuntimeClassInfoImpl.RuntimePropertySeed(super.createFieldSeed(field), acc);
    }
    
    public RuntimeClassInfoImpl.RuntimePropertySeed createAccessorSeed(Method getter, Method setter) {
        Accessor acc;
        try {
            acc = this.accessorFactory.createPropertyAccessor((Class)this.clazz, getter, setter);
        }
        catch (JAXBException e) {
            this.builder.reportError(new IllegalAnnotationException(Messages.CUSTOM_ACCESSORFACTORY_PROPERTY_ERROR.format(new Object[] { this.nav().getClassName(this.clazz), e.toString() }), this));
            acc = Accessor.getErrorInstance();
        }
        return new RuntimeClassInfoImpl.RuntimePropertySeed(super.createAccessorSeed(getter, setter), acc);
    }
    
    @Override
    protected void checkFieldXmlLocation(Field f) {
        if (this.reader().hasFieldAnnotation(XmlLocation.class, f)) {
            this.xmlLocationAccessor = new Accessor.FieldReflection(f);
        }
    }
    
    public Accessor<?, Locator> getLocatorField() {
        return this.xmlLocationAccessor;
    }
    
    static final class RuntimePropertySeed implements PropertySeed<Type, Class, Field, Method>
    {
        private final Accessor acc;
        private final PropertySeed<Type, Class, Field, Method> core;
        
        public RuntimePropertySeed(PropertySeed<Type, Class, Field, Method> core, Accessor acc) {
            this.core = core;
            this.acc = acc;
        }
        
        @Override
        public String getName() {
            return this.core.getName();
        }
        
        public <A extends Annotation> A readAnnotation(Class<A> annotationType) {
            return this.core.readAnnotation(annotationType);
        }
        
        public boolean hasAnnotation(Class<? extends Annotation> annotationType) {
            return this.core.hasAnnotation(annotationType);
        }
        
        @Override
        public Type getRawType() {
            return (Type)this.core.getRawType();
        }
        
        public Location getLocation() {
            return this.core.getLocation();
        }
        
        public Locatable getUpstream() {
            return this.core.getUpstream();
        }
        
        public Accessor getAccessor() {
            return this.acc;
        }
    }
    
    private static final class TransducerImpl<BeanT> implements Transducer<BeanT>
    {
        private final TransducedAccessor<BeanT> xacc;
        private final Class<BeanT> ownerClass;
        
        public TransducerImpl(Class<BeanT> ownerClass, TransducedAccessor<BeanT> xacc) {
            this.xacc = xacc;
            this.ownerClass = ownerClass;
        }
        
        public boolean useNamespace() {
            return this.xacc.useNamespace();
        }
        
        public boolean isDefault() {
            return false;
        }
        
        public void declareNamespace(BeanT bean, XMLSerializer w) throws AccessorException {
            try {
                this.xacc.declareNamespace(bean, w);
            }
            catch (SAXException e) {
                throw new AccessorException(e);
            }
        }
        
        @NotNull
        public CharSequence print(BeanT o) throws AccessorException {
            try {
                CharSequence value = this.xacc.print(o);
                if (value == null) {
                    throw new AccessorException(Messages.THERE_MUST_BE_VALUE_IN_XMLVALUE.format(new Object[] { o }));
                }
                return value;
            }
            catch (SAXException e) {
                throw new AccessorException(e);
            }
        }
        
        public BeanT parse(CharSequence lexical) throws AccessorException, SAXException {
            UnmarshallingContext ctxt = UnmarshallingContext.getInstance();
            BeanT inst;
            if (ctxt != null) {
                inst = ctxt.createInstance(this.ownerClass);
            }
            else {
                inst = ClassFactory.create(this.ownerClass);
            }
            this.xacc.parse(inst, lexical);
            return inst;
        }
        
        public void writeText(XMLSerializer w, BeanT o, String fieldName) throws IOException, SAXException, XMLStreamException, AccessorException {
            if (!this.xacc.hasValue(o)) {
                throw new AccessorException(Messages.THERE_MUST_BE_VALUE_IN_XMLVALUE.format(new Object[] { o }));
            }
            this.xacc.writeText(w, o, fieldName);
            return;
        }
        
        public void writeLeafElement(XMLSerializer w, Name tagName, BeanT o, String fieldName) throws IOException, SAXException, XMLStreamException, AccessorException {
            if (!this.xacc.hasValue(o)) {
                throw new AccessorException(Messages.THERE_MUST_BE_VALUE_IN_XMLVALUE.format(new Object[] { o }));
            }
            this.xacc.writeLeafElement(w, tagName, o, fieldName);
            return;
        }
        
        public QName getTypeName(BeanT instance) {
            return null;
        }
    }
}
